"""The matplotlib plot specific module."""

import queue
import time
from functools import partial
from typing import TYPE_CHECKING, Any, Generator

import matplotlib.pyplot as plt  # type: ignore
from matplotlib import _pylab_helpers
from matplotlib.animation import FuncAnimation  # type: ignore
from matplotlib.animation import FFMpegWriter, PillowWriter
from nxscli.idata import PluginData, PluginDataCb, PluginQueueData
from nxscli.logger import logger

if TYPE_CHECKING:
    from matplotlib.axes import Axes  # type: ignore
    from matplotlib.figure import Figure  # type: ignore
    from matplotlib.lines import Line2D  # type: ignore
    from nxscli.trigger import TriggerHandler
    from nxslib.dev import DeviceChannel
    from nxslib.nxscope import DNxscopeStream


###############################################################################
# Class: MplManager
###############################################################################


class MplManager:
    """Matplotlib global manager."""

    @staticmethod
    def draw() -> None:  # pragma: no cover
        """Draw an animation."""
        plt.draw()

    @staticmethod
    def fig_is_open() -> Any:  # pragma: no cover
        """Return a list of opened figures."""
        return plt.get_fignums()

    @staticmethod
    def pause(interval: float) -> None:  # pragma: no cover
        """Handle Matplotlib events.

        Modified pyplot.pause() without show(False) in the middle.

        :param interval: pause interval
        """
        manager = _pylab_helpers.Gcf.get_active()
        if manager is not None:
            canvas = manager.canvas
            if canvas.figure.stale:
                canvas.draw_idle()
            # show(block=False)
            canvas.start_event_loop(interval)
        else:
            time.sleep(interval)

    @staticmethod
    def show(block: bool = True) -> None:
        """Show an animation.

        :param block: blocking operation
        """
        plt.show(block=block)  # pragma: no cover

    @staticmethod
    def style_set(style: list[str]) -> None:
        """Configure matplotlib."""
        logger.info("plt.style %s", str(style))
        plt.style.use(style)

    @staticmethod
    def figure(dpi: float = 100.0) -> "Figure":
        """Get figure.

        :param dpi: figure DPI
        """
        return plt.figure(dpi=dpi)

    @staticmethod
    def func_animation(**kwargs: Any) -> Any:
        """Create animation.

        :param kwargs: animation arugments
        """
        return FuncAnimation(**kwargs)

    @staticmethod
    def close(fig: "Figure") -> None:
        """Close figure.

        :param fig: matplotlib Figure
        """
        if plt.fignum_exists(fig.number):  # pragma: no cover
            plt.close(fig)


###############################################################################
# Class: PlotDataCommon
###############################################################################


class PlotDataCommon:
    """A class implementing common plot data."""

    def __init__(self, channel: "DeviceChannel"):
        """Initialize common plot data.

        :param channel: channel instance
        """
        self._xdata: list[Any] = []
        self._ydata: list[Any] = []
        self._vdim = channel.data.vdim
        self._chan = channel.data.chan
        for _ in range(self._vdim):
            self._xdata.append([])
            self._ydata.append([])

        self._samples_max = 0

    @property
    def chan(self) -> int:
        """Get channel id."""
        return self._chan

    @property
    def xdata(self) -> list[list[Any]]:
        """Get X data."""
        return self._xdata

    @property
    def ydata(self) -> list[list[Any]]:
        """Get Y data."""
        return self._ydata

    @property
    def samples_max(self) -> int:
        """Get max samples."""
        return self._samples_max

    @samples_max.setter
    def samples_max(self, smax: int) -> None:
        """Set max samples.

        :param smax: set max num of samples
        """
        self._samples_max = smax

    def xdata_extend(self, data: list[list[Any]]) -> None:
        """Extend X data.

        :param data: X data to extend
        """
        for i, xdata in enumerate(self._xdata):
            xdata.extend(data[i])

    def ydata_extend(self, data: list[list[Any]]) -> None:
        """Extend Y data.

        :param data: Y data to extend
        """
        for i, ydata in enumerate(self._ydata):
            ydata.extend(data[i])

    def xdata_extend_max(self, data: list[list[Any]]) -> None:
        """Extend X data and saturate to a configured number of samples.

        :param data: X data to extend
        """
        for i, _ in enumerate(self._xdata):
            self._xdata[i].extend(data[i])
            remove = len(self._xdata[i]) - self._samples_max
            if remove > 0:
                self._xdata[i] = self._xdata[i][remove:]

    def ydata_extend_max(self, data: list[list[Any]]) -> None:
        """Extend Y data and saturate to a configured number of samples.

        :param data: Y data to extend
        """
        for i, _ in enumerate(self._xdata):
            self._ydata[i].extend(data[i])
            remove = len(self._ydata[i]) - self._samples_max
            if remove > 0:
                self._ydata[i] = self._ydata[i][remove:]


###############################################################################
# Class: PlotDataAxesMpl
###############################################################################


class PlotDataAxesMpl(PlotDataCommon):
    """A class implementing common matplotlib axes logic."""

    def __init__(
        self,
        ax: "Axes",
        channel: "DeviceChannel",
        fmt: list[str] | None = None,
    ):
        """Initialize matplotlib specific plot data.

        :param ax: matplotlib Axes
        :param channel: channel instance
        :param fmt: plot format
        """
        PlotDataCommon.__init__(self, channel)

        # initialize axis only if numerical channel
        if not channel.data.is_numerical:
            raise TypeError

        self._ax = ax

        if not fmt:
            # extend for all vectors
            self._fmt = ["" for _ in range(channel.data.vdim)]
        else:
            # we have to configure each vector individualy
            assert (
                len(fmt) == channel.data.vdim
            ), "fmt must match vectors in configured channel"
            self._fmt = fmt

        # we need lines for animations
        self._lns: list["Line2D"] = []
        for i in range(channel.data.vdim):
            lines = self._ax.plot([], [], self._fmt[i])
            self._lns.append(lines[0])

        # set grid
        self.grid_set(True)

        # set plot title if channel name available
        if len(channel.data.name) > 0:  # pragma: no cover
            self.plot_title = channel.data.name

    def __str__(self) -> str:
        """Format string representation."""
        _str = "PlotDataAxesMpl" + "(channel=" + str(self.chan) + ")"
        return _str

    @property
    def ax(self) -> "Axes":
        """Get axes."""
        return self._ax

    @property
    def lns(self) -> list["Line2D"]:
        """Get lines."""
        return self._lns

    @property
    def xlim(self) -> Any:
        """Get pot X limits."""
        assert self._ax
        return self._ax.get_xlim()

    @property
    def ylim(self) -> Any:
        """Get pot Y limits."""
        assert self._ax
        return self._ax.get_ylim()

    @property
    def plot_title(self) -> Any:
        """Get the plot title."""
        assert self._ax
        return self._ax.get_title()

    @plot_title.setter
    def plot_title(self, title: str) -> None:
        """Set the plot title.

        :param title: plot title
        """
        assert self._ax
        self._ax.set_title(title)

    def set_xlim(self, xlim: tuple[Any, Any]) -> None:
        """Set plot X limits.

        :param xlim: set X limits
        """
        assert self._ax
        self._ax.set_xlim(*xlim)

    def set_ylim(self, ylim: tuple[Any, Any]) -> None:
        """Set plot Y limits.

        :param ylim: set Y limits
        """
        assert self._ax
        self._ax.set_ylim(*ylim)

    def plot(self) -> None:
        """Plot all data."""
        assert self._ax
        for i, data in enumerate(self._ydata):
            self._ax.plot(data, self._fmt[i])

    def xaxis_disable(self) -> None:
        """Disable x axis ticks."""
        self.xaxis_set_ticks([])

    def xaxis_set_ticks(self, ticks: list[Any]) -> None:
        """Set ticks for X axis.

        :param ticks: set ticks for X axis
        """
        assert self._ax
        self._ax.get_xaxis().set_ticks(ticks)

    def grid_set(self, enable: bool) -> None:
        """Enable grid on plots.

        :param enable: enable or disable grid
        """
        assert self._ax
        self._ax.grid(enable)


###############################################################################
# Class: PluginAnimationCommonMpl
###############################################################################


class PluginAnimationCommonMpl:
    """A class implementing a common matplotlib animation plot logic."""

    def __init__(
        self,
        fig: "Figure",
        pdata: PlotDataAxesMpl,
        qdata: PluginQueueData,
        write: str,
    ):
        """Initialize animation handler.

        :param fig: matplotlib Figure
        :param pdata: axes handler
        :param qdata: stream queue handler
        :param kwargs: implementation specific arguments
        """
        self._fig = fig
        self._cnt = 0
        self._pdata = pdata
        self._qdata = qdata
        self._ani = None

        if write:  # pragma: no cover
            fps = 10
            logger.info("writer animation to file=%s, fps=%d", write, fps)

            tmp = write.split(".")
            if tmp[-1] == "gif":
                self._writer = PillowWriter(fps=fps)
            elif tmp[-1] == "mp4":
                bitrate = 200
                self._writer = FFMpegWriter(fps=fps, bitrate=bitrate)
            else:
                raise TypeError

            # NOTE: dpi arg set cause overlapping animation
            self._writer.setup(self._fig, write)
        else:
            self._writer = None

    def _animation_init(self, pdata: PlotDataAxesMpl) -> "Line2D":
        return pdata.lns

    def _animation_frames(
        self, qdata: PluginQueueData
    ) -> Generator[Any, None, None]:  # pragma: no cover
        ydata: list[list[Any]] = []
        xdata: list[list[Any]] = []

        for _ in range(self._qdata.vdim):
            ydata.append([])
            xdata.append([])

        # limit to 100 samples per frame
        for _ in range(100):
            data: list["DNxscopeStream"] = []
            try:
                # this must be non-blocking for queue.Empty exception
                data = qdata.queue_get(block=False)
            except queue.Empty:
                break

            # print("qsize=", qdata._queue.qsize())
            for sample in data:
                for i in range(self._qdata.vdim):
                    ydata[i].append(sample.data[i])
                    xdata[i].append(self._cnt)
                self._cnt += 1

        yield xdata, ydata

    def _animation_update(
        self, frame: tuple[list[Any], list[Any]], pdata: PlotDataAxesMpl
    ) -> list["Line2D"] | None:
        pass  # pragma: no cover

    def pause(self) -> None:
        """Pause an animation."""
        if self._ani is not None:  # pragma: no cover
            self._ani.pause()

    def stop(self) -> None:  # pragma: no cover
        """Stop animation."""
        # TODO: stop animation
        if self._ani:
            if self._ani.event_source:
                self._ani.pause()
        if self._writer:
            self._writer.finish()

        del self._ani

    def _animation_update_cmn(
        self, frame: tuple[list[Any], list[Any]], pdata: PlotDataAxesMpl
    ) -> list["Line2D"]:  # pragma: no cover
        """Update animation common logic."""
        # no data
        if len(frame[0]) == 0 or len(frame[1]) == 0:
            return pdata.lns

        # implementation specific
        lines = self._animation_update(frame, pdata)
        assert lines

        # handle writer
        if self._writer:
            # self._fig.canvas.flush_events()
            self._writer.grab_frame()

        return lines

    def start(self) -> None:
        """Start an animation."""
        fig = self._fig
        update = partial(self._animation_update_cmn, pdata=self._pdata)
        frames = partial(self._animation_frames, qdata=self._qdata)
        init = partial(self._animation_init, pdata=self._pdata)
        self._ani = MplManager.func_animation(
            fig=fig,
            func=update,
            frames=frames,
            init_func=init,
            interval=1,
            blit=True,
            cache_frame_data=False,
        )

    def xaxis_disable(self) -> None:  # pragma: no cover
        """Hide x axis."""
        self._pdata.xaxis_disable()

    def yscale_extend(
        self, frame: list[Any], pdata: PlotDataAxesMpl, scale: float = 1.1
    ) -> None:  # pragma: no cover
        """Extend yscale if needed with a given scale factor.

        :param frame: frame data
        :param pdata: axes handler
        :param scale: scale factor
        """
        assert pdata.ax
        ymin, ymax = pdata.ax.get_ylim()

        new_ymax = ymax
        new_ymin = ymin
        for data in frame:
            # do nothing if empty
            if not data:
                return

            # get min/max
            ytmp = max(data)
            if ytmp > new_ymax:
                new_ymax = ytmp
            ytmp = min(data)
            if ytmp < new_ymin:
                new_ymin = ytmp

        if new_ymax > ymax:
            new_ymax = new_ymax * scale
            ymax = new_ymax  # store for ymin update
            pdata.ax.set_ylim(ymin, new_ymax)
            pdata.ax.figure.canvas.draw()
        if new_ymin < ymin:
            new_ymin = new_ymin * scale
            pdata.ax.set_ylim(new_ymin, ymax)
            pdata.ax.figure.canvas.draw()

    def xscale_extend(
        self, frame: list[Any], pdata: PlotDataAxesMpl, scale: float = 2.0
    ) -> None:  # pragma: no cover
        """Exten x axis if needed with a agiven scale factor.

        :param frame: frame data
        :param pdata: axes handler
        :param scale: scale factor
        """
        assert pdata.ax
        xmin, xmax = pdata.ax.get_xlim()

        tmax = xmax
        # get min/max
        for i in frame:
            if len(i) > 0:
                tmp = max(i)
                if tmp > tmax:
                    tmax = tmp

        # change x scale
        if tmax > xmax:
            pdata.ax.set_xlim(xmin, scale * xmax)
            pdata.ax.figure.canvas.draw()

    def xscale_saturate(
        self, _: list[Any], pdata: PlotDataAxesMpl
    ) -> None:  # pragma: no cover
        """Saturate x axis.

        :param pdata: axes handler
        """
        assert pdata.ax
        xmin, xmax = pdata.ax.get_xlim()

        # change x scale fit for xdata
        new_xmin = pdata.xdata[0]
        new_xmax = pdata.xdata[-1]
        if xmin > new_xmin or xmax < new_xmax:
            pdata.ax.set_xlim(new_xmin, new_xmax)
            # TODO: revisit
            # pdata.ax.figure.canvas.draw()


###############################################################################
# Class: PluginPlotMpl
###############################################################################


class PluginPlotMpl(PluginData):
    """A class implementing matplotlib common plot handler."""

    def __init__(
        self,
        chanlist: list["DeviceChannel"],
        trig: list["TriggerHandler"],
        cb: PluginDataCb,
        dpi: float = 100.0,
        fmt: list[str] | None = None,
    ):
        """Intiialize a plot handler.

        :param chanlist: a list with plugin channels
        :param cb: plugin callback to nxslib
        :param dpi: figure DPI
        :param fmt: plot format
        """
        logger.info("prepare plot %s", str(chanlist))
        newchanlist = []
        for chan in chanlist:
            # get only numerical channels
            if chan.data.is_numerical:
                newchanlist.append(chan)
            else:  # pragma: no cover
                logger.info(
                    "NOTE: channel %d not numerical - ignore", chan.data.chan
                )
        assert len(newchanlist) == len(trig)

        super().__init__(newchanlist, trig, cb)

        self._fig = MplManager.figure(dpi)
        self._ax: list[Axes] = []
        self._ani: list[PluginAnimationCommonMpl] = []

        self._fmt: list[Any]
        if not fmt:
            # defaul configuration for all
            self._fmt = [None for _ in range(len(self._chanlist))]
        elif len(self._chanlist) != 1 and len(fmt) == 1:
            # the same format for all channels - extend fmt for all channels
            self._fmt = [
                [fmt[0]] * self._chanlist[i].data.vdim
                for i in range(len(self._chanlist))
            ]
        else:
            # individual fmt for all channels
            assert len(fmt) == len(
                self._chanlist
            ), "fmt must be specified for all configured channels"
            self._fmt = fmt

        # add subplots
        self._add_subplots(newchanlist)

        self._plist = self._plist_init()

    def __del__(self) -> None:
        """Close figure and clean queue handlers."""
        MplManager.close(self._fig)
        super().__del__()

    def _plist_init(self) -> list[PlotDataAxesMpl]:
        ret = []
        for i, channel in enumerate(self._chanlist):
            logger.info(
                "intialize PlotDataAxesMpl chan=%d vdim=%d fmt=%s",
                channel.data.chan,
                channel.data.vdim,
                self._fmt[i],
            )
            # initialize plot
            pdata = PlotDataAxesMpl(self._ax[i], channel, fmt=self._fmt[i])
            # add plot to list
            ret.append(pdata)
        return ret

    def _add_subplots(self, channels: list["DeviceChannel"]) -> None:
        """Create subplots from channels list."""
        # remove all current axes
        for ax in self._ax:  # pragma: no cover
            if ax is not None:
                self._fig.delaxes(ax)
                # ax.remove()
            else:
                pass
        self._ax = []

        # show plots only for numerical channels
        chanlist = []
        for chan in channels:
            if chan.data.is_numerical:
                chanlist.append(chan.data.chan)
            else:  # pragma: no cover
                pass

        row = len(chanlist)
        col = 1

        i = 1
        for _ in range(len(chanlist)):
            # create subplot for all used numerical channels
            self._ax.append(self._fig.add_subplot(row, col, i))
            i += 1

    @property
    def fig(self) -> "Figure":
        """Get figure handler."""
        return self._fig

    @property
    def ani(self) -> list[PluginAnimationCommonMpl]:
        """Return all registered animation isntances."""
        return self._ani

    @property
    def plist(self) -> list[PlotDataAxesMpl]:
        """Get plotdata list."""
        return self._plist

    def ani_append(self, ani: PluginAnimationCommonMpl) -> None:
        """Add animation.

        :param ani: plugin animation handler
        """
        self._ani.append(ani)

    def ani_clear(self) -> None:  # pragma: no cover
        """Clear animations."""
        # TODO: fix me, doesnt work
        del self._ani
        self._ani = []

    def plot_clear(self) -> None:
        """Clear plot data."""
        if len(self._ax) > 0:  # pragma: no cover
            for ax in self._ax:
                if ax is not None:
                    ax.cla()



"""Module containing animation1 plugin command."""

from typing import TYPE_CHECKING

import click
from nxscli.cli.environment import Environment, pass_environment

from nxscli_mpl.cli.types import plot_options

if TYPE_CHECKING:
    from nxscli.trigger import DTriggerConfigReq


###############################################################################
# Command: cmd_pani1
###############################################################################


@click.command(name="pani1")
@plot_options
@pass_environment
def cmd_pani1(
    ctx: Environment,
    chan: list[int],
    trig: dict[int, "DTriggerConfigReq"],
    dpi: float,
    fmt: list[list[str]],
    write: str,
) -> bool:
    """[plugin] Animation plot without a length limit (infinite plot)."""
    assert ctx.phandler
    ctx.phandler.enable(
        "animation1", channels=chan, trig=trig, dpi=dpi, fmt=fmt, write=write
    )

    ctx.needchannels = True

    return True



"""Module containing animation1 plugin."""

from typing import TYPE_CHECKING, Any

from nxscli_mpl.animation_mpl import IPluginAnimation
from nxscli_mpl.plot_mpl import PlotDataAxesMpl, PluginAnimationCommonMpl

if TYPE_CHECKING:
    from matplotlib.figure import Figure  # type: ignore
    from matplotlib.lines import Line2D  # type: ignore
    from nxscli.idata import PluginQueueData


###############################################################################
# Class: Animation1
###############################################################################


class Animation1(PluginAnimationCommonMpl):
    """Infinity animation with x axis extension."""

    def __init__(
        self,
        fig: "Figure",
        pdata: PlotDataAxesMpl,
        qdata: "PluginQueueData",
        write: str,
    ) -> None:
        """Initialzie an animtaion1 handler.

        :param fig: matplotlib Figure
        :param pdata: axes handler
        :param qdata: stream queue handler
        :param write: write path
        """
        PluginAnimationCommonMpl.__init__(self, fig, pdata, qdata, write)

    def _animation_update(
        self, frame: tuple[list[Any], list[Any]], pdata: PlotDataAxesMpl
    ) -> "Line2D":  # pragma: no cover
        """Update an animation with dynamic scaling."""
        # update sample
        pdata.xdata_extend(frame[0])
        pdata.ydata_extend(frame[1])

        # update y scale
        self.yscale_extend(frame[1], pdata)

        # update x scale
        self.xscale_extend(frame[0], pdata)

        # set new data
        i = 0
        for ln in pdata.lns:
            ln.set_data(pdata.xdata[i], pdata.ydata[i])
            i += 1

        return pdata.lns


###############################################################################
# Class: PluginAnimation1
###############################################################################


class PluginAnimation1(IPluginAnimation):
    """Infinity animation with x axis extension."""

    def __init__(self) -> None:
        """Initialize an animation1 plugin."""
        IPluginAnimation.__init__(self)

    def _start(
        self,
        fig: "Figure",
        pdata: PlotDataAxesMpl,
        qdata: "PluginQueueData",
        kwargs: Any,
    ) -> PluginAnimationCommonMpl:
        """Start an animation1 plugin."""
        return Animation1(fig, pdata, qdata, kwargs["write"])



import queue

import pytest  # type: ignore
from matplotlib.axes import Axes  # type: ignore
from matplotlib.figure import Figure  # type: ignore
from nxscli.idata import PluginDataCb, PluginQueueData
from nxscli.trigger import DTriggerConfig, ETriggerType, TriggerHandler
from nxslib.dev import DeviceChannel

from nxscli_mpl.plot_mpl import (
    PlotDataAxesMpl,
    PlotDataCommon,
    PluginAnimationCommonMpl,
    PluginPlotMpl,
)


def test_plotdatacommon():
    chan = DeviceChannel(0, 1, 2, "chan0")
    x = PlotDataCommon(chan)

    assert x.samples_max == 0
    x.samples_max = 100
    assert x.samples_max == 100

    assert x.xdata == [[], []]
    assert x.ydata == [[], []]

    x.xdata_extend([[1, 2], [3, 4]])
    x.ydata_extend([[5, 6], [7, 8]])
    assert x.xdata == [[1, 2], [3, 4]]
    assert x.ydata == [[5, 6], [7, 8]]
    x.xdata_extend([[9], [10]])
    x.ydata_extend([[11], [12]])
    assert x.xdata == [[1, 2, 9], [3, 4, 10]]
    assert x.ydata == [[5, 6, 11], [7, 8, 12]]

    x.samples_max = 5
    x.xdata_extend_max([[13, 14], [16, 17]])
    x.ydata_extend_max([[19, 20], [22, 23]])
    x.xdata_extend_max([[15], [18]])
    x.ydata_extend_max([[21], [24]])

    assert x.xdata == [[2, 9, 13, 14, 15], [4, 10, 16, 17, 18]]
    assert x.ydata == [[6, 11, 19, 20, 21], [8, 12, 22, 23, 24]]
    x.xdata_extend_max([[25], [26]])
    x.ydata_extend_max([[27], [28]])
    assert x.xdata == [[9, 13, 14, 15, 25], [10, 16, 17, 18, 26]]
    assert x.ydata == [[11, 19, 20, 21, 27], [12, 22, 23, 24, 28]]


def test_plotdataaxesmpl():
    fig = Figure()
    axes = Axes(fig, (1, 1, 2, 6))

    # not numerical channels
    chan = DeviceChannel(0, 1, 2, "chan0")
    with pytest.raises(TypeError):
        x = PlotDataAxesMpl(axes, chan)

    chan = DeviceChannel(chan=0, _type=2, vdim=2, name="chan0")
    x = PlotDataAxesMpl(axes, chan)

    assert x.ax is axes
    assert str(x) is not None

    x.set_xlim((0, 1))
    assert x.xlim == (0, 1)
    x.set_ylim((2, 3))
    assert x.ylim == (2, 3)

    x.plot_title = "test"
    assert x.plot_title == "test"

    x.plot()
    x.xaxis_disable()
    x.xaxis_set_ticks([])

    x = PlotDataAxesMpl(axes, chan, fmt=None)
    assert x._fmt == ["", ""]

    x = PlotDataAxesMpl(axes, chan, fmt=["o", "b"])
    assert x._fmt == ["o", "b"]


def test_pluginanimationcommonmpl():
    q = queue.Queue()
    chan = DeviceChannel(chan=0, _type=2, vdim=2, name="chan0")
    fig = Figure()
    axes = Axes(fig, (1, 1, 2, 6))
    pdata = PlotDataAxesMpl(axes, chan)
    dtc = DTriggerConfig(ETriggerType.ALWAYS_OFF)
    qdata = PluginQueueData(q, chan, dtc)
    x = PluginAnimationCommonMpl(fig, pdata, qdata, False)

    x.stop()
    # TODO


def dummy_stream_sub(ch):
    pass


def dummy_stream_unsub(q):
    pass


def test_pluginplotmpl():
    chanlist = [
        DeviceChannel(chan=0, _type=1, vdim=2, name="chan0"),  # not numerical
        DeviceChannel(chan=1, _type=2, vdim=1, name="chan1"),
        DeviceChannel(chan=2, _type=2, vdim=2, name="chan2"),
    ]
    dtc = DTriggerConfig(ETriggerType.ALWAYS_OFF)
    trig = [TriggerHandler(1, dtc), TriggerHandler(2, dtc)]
    cb = PluginDataCb(dummy_stream_sub, dummy_stream_unsub)
    x = PluginPlotMpl(chanlist, trig, cb)

    assert x.fig is not None
    assert x.ani == []
    assert len(x.plist) > 0
    assert len(x._chanlist) == 2  # one channel not numerical
    assert x._fmt == [None, None]

    # test fmt configuration
    x = PluginPlotMpl(chanlist, trig, cb, fmt="o")
    assert x._fmt == [["o"], ["o", "o"]]

    x = PluginPlotMpl(chanlist, trig, cb, fmt=[["o"], ["b", "b"]])
    assert x._fmt == [["o"], ["b", "b"]]

    # invalid vector fmt for chan 2
    with pytest.raises(AssertionError):
        x = PluginPlotMpl(chanlist, trig, cb, fmt=["o", "b"])
    # invalid channels fmt
    with pytest.raises(AssertionError):
        x = PluginPlotMpl(chanlist, trig, cb, fmt=["o", "b", "c"])

    # TODO

    TriggerHandler.cls_cleanup()



"""Module containing the common matplotlib animation plugin logic."""

from abc import abstractmethod
from typing import TYPE_CHECKING, Any

from nxscli.iplugin import IPluginPlotDynamic
from nxscli.logger import logger

from nxscli_mpl.plot_mpl import (
    MplManager,
    PlotDataAxesMpl,
    PluginAnimationCommonMpl,
    PluginPlotMpl,
)

if TYPE_CHECKING:
    from matplotlib.figure import Figure  # type: ignore
    from nxscli.idata import PluginQueueData


###############################################################################
# Class: IPluginAnimation
###############################################################################


class IPluginAnimation(IPluginPlotDynamic):
    """The common logic for an animation plugin."""

    def __init__(self) -> None:
        """Initialize an animation plugin."""
        super().__init__()

        self._plot: "PluginPlotMpl"

    @abstractmethod
    def _start(
        self,
        fig: "Figure",
        pdata: "PlotDataAxesMpl",
        qdata: "PluginQueueData",
        kwargs: Any,
    ) -> "PluginAnimationCommonMpl":
        """Abstract method.

        :param fig: matplotlib Figure
        :param pdata: axes handler
        :param qdata: stream queue handler
        :param kwargs: implementation specific arguments
        """

    @property
    def stream(self) -> bool:
        """Return True if this plugin needs stream."""
        return True

    def wait_for_plugin(self) -> bool:  # pragma: no cover
        """Wait for figure to close."""
        done = True
        if MplManager.fig_is_open():
            done = False
            # pause
            MplManager.pause(1)
        return done

    def stop(self) -> None:
        """Stop all animations."""
        assert self._plot

        if len(self._plot.ani) > 0:
            for ani in self._plot.ani:
                ani.stop()

    def clear(self) -> None:
        """Clear all animations."""
        assert self._plot

        self._plot.ani_clear()

    def data_wait(self, timeout: float = 0.0) -> bool:
        """Return True if data are ready.

        :param timeout: data wait timeout
        """
        return True

    def start(self, kwargs: Any) -> bool:
        """Start animation plugin.

        :param kwargs: implementation specific arguments
        """
        assert self._phandler

        logger.info("start %s", str(kwargs))

        chanlist = self._phandler.chanlist_plugin(kwargs["channels"])
        trig = self._phandler.triggers_plugin(chanlist, kwargs["trig"])

        cb = self._phandler.cb_get()
        self._plot = PluginPlotMpl(
            chanlist, trig, cb, dpi=kwargs["dpi"], fmt=kwargs["fmt"]
        )

        # clear previous animations
        self.clear()

        # new animations
        for i, pdata in enumerate(self._plot.plist):
            ani = self._start(
                self._plot.fig, pdata, self._plot.qdlist[i], kwargs
            )
            self._plot.ani_append(ani)

        for ani in self._plot.ani:
            ani.start()

        return True

    def result(self) -> "PluginPlotMpl":
        """Get animation plugin result."""
        assert self._plot
        MplManager.show(block=False)
        return self._plot



"""Module containing capture plugin."""

from typing import TYPE_CHECKING, Any

from nxscli.iplugin import IPluginPlotStatic
from nxscli.logger import logger
from nxscli.pluginthr import PluginThread

from nxscli_mpl.plot_mpl import MplManager, PluginPlotMpl

if TYPE_CHECKING:
    from nxscli.idata import PluginQueueData
    from nxslib.nxscope import DNxscopeStream


###############################################################################
# Class: PluginCapture
###############################################################################


class PluginCapture(PluginThread, IPluginPlotStatic):
    """Plugin that plot static captured data."""

    def __init__(self) -> None:
        """Intiialize a capture plot plugin."""
        IPluginPlotStatic.__init__(self)
        PluginThread.__init__(self)

        self._plot: "PluginPlotMpl"
        self._write: str

    def _init(self) -> None:
        assert self._phandler

    def _final(self) -> None:
        logger.info("plot capture DONE")

    def _handle_samples(
        self, data: list["DNxscopeStream"], pdata: "PluginQueueData", j: int
    ) -> None:
        # store data
        ydata: list[list[Any]] = [[] for v in range(pdata.vdim)]
        for sample in data:
            for i in range(pdata.vdim):
                # TODO: metadata not supported for now
                ydata[i].append(sample.data[i])

        # extend ydata
        self._plot.plist[j].ydata_extend(ydata)
        # get data len
        self._datalen[j] = len(self._plot.plist[j].ydata[0])

    def wait_for_plugin(self) -> bool:  # pragma: no cover
        """Wait for figure to close."""
        done = True
        if MplManager.fig_is_open():
            done = False
            # pause
            MplManager.pause(1)
        return done

    def start(self, kwargs: Any) -> bool:
        """Start capture plugin.

        :param kwargs: implementation specific arguments
        """
        assert self._phandler

        logger.info("start capture %s", str(kwargs))

        self._samples = kwargs["samples"]
        self._write = kwargs["write"]
        self._nostop = kwargs["nostop"]

        chanlist = self._phandler.chanlist_plugin(kwargs["channels"])
        trig = self._phandler.triggers_plugin(chanlist, kwargs["trig"])

        cb = self._phandler.cb_get()
        self._plot = PluginPlotMpl(
            chanlist, trig, cb, dpi=kwargs["dpi"], fmt=kwargs["fmt"]
        )

        if not self._plot.qdlist or not self._plot.plist:  # pragma: no cover
            return False

        for pdata in self._plot.plist:
            # update xlim to fit our data
            if self._samples:
                pdata.set_xlim((0, self._samples))
            else:  # pragma: no cover
                pass

        self.thread_start(self._plot)

        return True

    def result(self) -> "PluginPlotMpl":
        """Get capture plugin result."""
        assert self._plot

        # plot all data
        for pdata in self._plot.plist:
            pdata.plot()

        if self._write:  # pragma: no cover
            self._plot.fig.savefig(self._write)

        MplManager.show(block=False)
        return self._plot



"""The matplotlib plot specific command."""

import click
from nxscli.cli.environment import Environment, pass_environment
from nxscli.cli.types import StringList

from nxscli_mpl.plot_mpl import MplManager

###############################################################################
# Command: cmd_mpl
###############################################################################


@click.command(name="mpl")
@click.option(
    "--style",
    default="ggplot,fast",
    type=StringList(),
    help="Configure Matplotlib style, default: ggplot, fast",
)
@pass_environment
def cmd_mpl(ctx: Environment, style: list[str]) -> bool:
    """[config] Matplotlib configuration."""  # noqa: D301
    MplManager.style_set(style)
    return True


# default configuration
MplManager.style_set(["ggplot", "fast"])



import nxscli_mpl
import nxscli_mpl.ext_commands
import nxscli_mpl.ext_plugins


def test_nxsclimpl():
    assert nxscli_mpl.__version__

    assert isinstance(nxscli_mpl.ext_plugins.plugins_list, list)
    assert isinstance(nxscli_mpl.ext_commands.commands_list, list)



"""Module containing animation2 plugin."""

from typing import TYPE_CHECKING, Any

from nxscli_mpl.animation_mpl import IPluginAnimation
from nxscli_mpl.plot_mpl import PlotDataAxesMpl, PluginAnimationCommonMpl

if TYPE_CHECKING:
    from matplotlib.figure import Figure  # type: ignore
    from matplotlib.lines import Line2D  # type: ignore
    from nxscli.idata import PluginQueueData


###############################################################################
# Class: Animation2
###############################################################################


class Animation2(PluginAnimationCommonMpl):
    """Animation with x axis saturation."""

    def __init__(
        self,
        fig: "Figure",
        pdata: PlotDataAxesMpl,
        qdata: "PluginQueueData",
        write: str,
        static_xticks: bool = True,
        disable_xaxis: bool = False,
    ):
        """Initialzie an animtaion2 handler.

        :param fig: matplotlib Figure
        :param pdata: axes handler
        :param qdata: stream queue handler
        :param write: write path
        :param static_xticks: use static X axis ticks
        :param disable_xaxis: disable X axis ticks
        """
        PluginAnimationCommonMpl.__init__(self, fig, pdata, qdata, write)

        if static_xticks is True:
            self._animation_update_priv = self._animation_update_staticx
        else:  # pragma: no cover
            self._animation_update_priv = self._animation_update_dynamicx

        if disable_xaxis is True:  # pragma: no cover
            self.xaxis_disable()

    def _animation_update(
        self, frame: tuple[list[Any], list[Any]], pdata: PlotDataAxesMpl
    ) -> "Line2D":  # pragma: no cover
        return self._animation_update_priv(frame, pdata)

    def _animation_update_staticx(
        self, frame: tuple[list[Any], list[Any]], pdata: PlotDataAxesMpl
    ) -> "Line2D":  # pragma: no cover
        """Update an animation with static X ticks."""
        # update sample
        pdata.xdata_extend_max(frame[0])
        pdata.ydata_extend_max(frame[1])

        # update y scale
        self.yscale_extend(frame[1], pdata)

        xdata = range(0, len(pdata.ydata[0]))
        i = 0
        for ln in pdata.lns:
            ln.set_data(xdata, pdata.ydata[i])
            i += 1

        return pdata.lns

    def _animation_update_dynamicx(
        self, frame: tuple[list[Any], list[Any]], pdata: PlotDataAxesMpl
    ) -> "Line2D":  # pragma: no cover
        """Update an animation with dynamic X ticks."""
        xdata = frame[0]
        ydata = frame[1]

        if not xdata or not ydata:
            return pdata.lns

        # update sample
        pdata.xdata_extend_max(xdata)
        pdata.ydata_extend_max(ydata)

        # update y scale
        self.yscale_extend(ydata, pdata)

        # update x scale
        self.xscale_saturate(xdata, pdata)

        # set new data
        i = 0
        for ln in pdata.lns:
            ln.set_data(pdata.xdata[i], pdata.ydata[i])
            i += 1

        return pdata.lns


###############################################################################
# Class: PluginAnimation2
###############################################################################


class PluginAnimation2(IPluginAnimation):
    """Animation with x axis saturation."""

    def __init__(self) -> None:
        """Initialize an animation2 plugin."""
        IPluginAnimation.__init__(self)

    def _start(
        self,
        fig: "Figure",
        pdata: PlotDataAxesMpl,
        qdata: "PluginQueueData",
        kwargs: Any,
    ) -> PluginAnimationCommonMpl:
        """Start an animation2 plugin."""
        maxsamples = kwargs["maxsamples"]

        # configure the max number of samples
        pdata.samples_max = maxsamples
        pdata.set_xlim((0, maxsamples))

        # start animation
        return Animation2(
            fig,
            pdata,
            qdata,
            kwargs["write"],
            static_xticks=True,
            disable_xaxis=False,
        )



import pytest  # type: ignore
from nxscli.phandler import PluginHandler
from nxslib.intf.dummy import DummyDev
from nxslib.nxscope import NxscopeHandler
from nxslib.proto.parse import Parser

from nxscli_mpl.animation_mpl import IPluginAnimation
from nxscli_mpl.plot_mpl import PluginAnimationCommonMpl


class XTestAnimation(PluginAnimationCommonMpl):
    def __init__(self, fig, pdata, qdata, write):
        super().__init__(fig, pdata, qdata, write)

    def _animation_update(self, frame, pdata):  # pragma: no cover
        pass


class XTestPluginAnimation(IPluginAnimation):
    def __init__(self):
        super().__init__()

    def _start(self, fig, pdata, qdata, kwargs):
        return XTestAnimation(fig, pdata, qdata, kwargs["write"])


def test_ipluginanimation_init():
    # abstract class
    with pytest.raises(TypeError):
        IPluginAnimation()

    x = XTestPluginAnimation()

    # phandler not connected
    with pytest.raises(AttributeError):
        x.start(None)
    with pytest.raises(AttributeError):
        x.result()
    with pytest.raises(AttributeError):
        x.clear()
    with pytest.raises(AttributeError):
        x.stop()

    assert x.stream is True
    assert x.data_wait() is True

    p = PluginHandler()
    x.connect_phandler(p)

    # clean up
    p.cleanup()


@pytest.fixture
def nxscope():
    intf = DummyDev()
    parse = Parser()
    nxscope = NxscopeHandler(intf, parse)
    return nxscope


def test_ipluginanimation_start_nochannels(nxscope):
    x = XTestPluginAnimation()
    p = PluginHandler()
    p.nxscope_connect(nxscope)
    x.connect_phandler(p)

    # start
    args = {"channels": [], "trig": [], "dpi": 100, "fmt": ""}
    assert x.start(args) is True

    # clear
    x.clear()

    # result
    x.result()

    # stop
    x.stop()

    # clean up
    p.cleanup()


def test_ipluginanimation_start(nxscope):
    x = XTestPluginAnimation()
    p = PluginHandler()
    p.nxscope_connect(nxscope)
    x.connect_phandler(p)

    # configure channels
    p.channels_configure([1], 0)

    # start
    args = {
        "channels": [1],
        "trig": [],
        "dpi": 100,
        "fmt": [""],
        "write": False,
    }
    assert x.start(args) is True

    # clear
    x.clear()

    # result
    x.result()

    # stop
    x.stop()

    # clean up
    p.cleanup()



"""Matplotlib extension commands."""

from typing import TYPE_CHECKING

from nxscli_mpl.commands.cmd_ani1 import cmd_pani1
from nxscli_mpl.commands.cmd_ani2 import cmd_pani2
from nxscli_mpl.commands.cmd_cap import cmd_pcap
from nxscli_mpl.commands.config.cmd_mpl import cmd_mpl

if TYPE_CHECKING:
    import click

commands_list: list["click.Command"] = [
    cmd_mpl,
    cmd_pcap,
    cmd_pani1,
    cmd_pani2,
]



"""Module containing animation2 plugin command."""

from typing import TYPE_CHECKING

import click
from nxscli.cli.environment import Environment, pass_environment

from nxscli_mpl.cli.types import plot_options

if TYPE_CHECKING:
    from nxscli.trigger import DTriggerConfigReq


###############################################################################
# Command: cmd_pani2
###############################################################################


@click.command(name="pani2")
@click.argument("maxsamples", type=int, required=True)
@plot_options
@pass_environment
def cmd_pani2(
    ctx: Environment,
    maxsamples: int,
    chan: list[int],
    trig: dict[int, "DTriggerConfigReq"],
    dpi: float,
    fmt: list[list[str]],
    write: str,
) -> bool:
    """[plugin] Animation plot with a lenght limit (saturated X-axis)."""
    assert ctx.phandler
    if maxsamples == 0:  # pragma: no cover
        click.secho("ERROR: Missing argument MAXSAMPLES", err=True, fg="red")
        return False

    ctx.phandler.enable(
        "animation2",
        maxsamples=maxsamples,
        channels=chan,
        trig=trig,
        dpi=dpi,
        fmt=fmt,
        write=write,
    )

    ctx.needchannels = True

    return True



"""Matplotlib extension plugins."""

from nxscli.iplugin import DPluginDescription

from nxscli_mpl.plugins.animation1 import PluginAnimation1
from nxscli_mpl.plugins.animation2 import PluginAnimation2
from nxscli_mpl.plugins.capture import PluginCapture

plugins_list = [
    DPluginDescription("capture", PluginCapture),
    DPluginDescription("animation1", PluginAnimation1),
    DPluginDescription("animation2", PluginAnimation2),
]



"""Module containing the Click types for nxscli-mpl."""

from typing import Any

import click
from nxscli.cli.types import (
    Channels,
    StringList2,
    Trigger,
    channels_option_help,
    trigger_option_help,
)

###############################################################################
# Globals: stirngs
###############################################################################


fmt_option_help = """Plugin specific Matplotlib format string configuration.
                     Channels separated by a semicolon (;),
                     vectors separated by a commas (?).
                     Example: 'r?g?b; -r?; r?b'
                     Defalut: Matplotlib default.
                      """  # noqa: D301


###############################################################################
# Decorator: plot_options
###############################################################################


# common plot options
_plot_options = (
    click.option(
        "--chan",
        default=None,
        type=Channels(),
        help=channels_option_help,
    ),
    click.option(
        "--trig",
        default=None,
        type=Trigger(),
        help=trigger_option_help,
    ),
    click.option("--dpi", type=int, default=100),
    click.option(
        "--fmt",
        default="",
        type=StringList2(ch1="?"),
        help=fmt_option_help,
    ),
    click.option("--write", type=click.Path(resolve_path=False), default=""),
)


def plot_options(fn: Any) -> Any:
    """Decorate command with common plot options decorator."""
    for decorator in reversed(_plot_options):
        fn = decorator(fn)
    return fn
